################################################################################
### function to find a next border that includes at least N new observations ###
################################################################################
next_value <- function(vec, border, N=1, lower=F){
  if(lower) {vec <- vec * (-1); border <- -border}
  sort_vec <- sort(vec)
  if(border > sort_vec[length(sort_vec)]) if(lower) return(-Inf) else return(Inf)
  if(!(border %in% vec)) border <- vec[which.min(ifelse(vec-border > 0, vec-border, Inf))]
  if(lower) res <- -sort_vec[sort_vec > border][N] else res <- sort_vec[sort_vec > border][N]
  if(lower & is.na(res)) res <- -Inf
  if(!lower & is.na(res)) res <- Inf
  return(res)
}

###########################################################################
### function that defines a box given a prim object and the step number ###
###########################################################################
define_fixbox <- function(prim, step){
  f <- prim$f[step]
  beta <- prim$beta[step]
  fixbox <- prim$box[step,]
  fixbox_metric <- prim$box_metric[step,]
  data <- list(all=prim$data_orig, box=prim$data[[step]])
  if(!is.null(prim$box_nom[[1]])) fixbox_nom <- prim$box_nom[[step]] else fixbox_nom <- NULL
  return(list(f=f, beta=beta, fixbox=fixbox, fixbox_metric=fixbox_metric, fixbox_nom=fixbox_nom, data=data))
}

#################################################################################
### function that returns the subset of the data that is contained in the box ###
#################################################################################
inbox_metric <- function(X, box) {
  subset(X, apply(sapply(2:ncol(X), function(k) X[,k] > box[,k-1] & X[,k] < box[,(ncol(X)-1)+k-1]),1,all ))
}
inbox_metric_index <- function(X, box) {
  apply(sapply(2:ncol(X), function(k) X[,k] > box[,k-1] & X[,k] < box[,(ncol(X)-1)+k-1]),1,all )
}
inbox_nom_index <- function(X, box){
  apply(sapply(1:length(box), function(k) X[, match(names(box)[k], names(X))]  %in% names(box[[k]])[as.vector(t(box[[k]]))]),1,all )
}
inbox <- function(X, fixbox_metric, fixbox_nom) {
  metric <- sapply(1:ncol(X), function(k) is.numeric(X[,k]) | is.ordered(X[,k]))
  if(sum(metric[-1])!=0) ind_metric <- inbox_metric_index(X[,metric], fixbox_metric) else ind_metric <- T
  if(sum(metric[-1])!=length(metric[-1])) ind_nom <- inbox_nom_index(X, fixbox_nom) else ind_nom <- T
  ind <- ind_metric & ind_nom
  X[ind,]  
} 

##########################################
### function to remove dominated boxes ###
##########################################
remove_dominated <- function(prim, sort=TRUE, dup.rm=TRUE){
  dominated<-sapply(1:length(prim$f), function(k) any(prim$f[-k] >= prim$f[k] & prim$beta[-k] >= prim$beta[k] &! (prim$f[-k] == prim$f[k] & prim$beta[-k] == prim$beta[k])))
  result <- list(f=prim$f[!dominated], beta=prim$beta[!dominated], data=prim$data[!dominated], box=prim$box[!dominated,], box_metric=prim$box_metric[!dominated,], box_nom=prim$box_nom[!dominated])
  if(dup.rm==T){
    dup <- duplicated(cbind(result$f,result$beta))
    result <- list(f=result$f[!dup], beta=result$beta[!dup], data=result$data[!dup], box=result$box[!dup,], box_metric=result$box_metric[!dup,], box_nom=result$box_nom[!dup])
  }
  if(sort==T){
    ord <- order(result$beta, decreasing = T)
    result$f <- result$f[ord]
    result$beta <- result$beta[ord]
    result$data <- result$data[ord]
    result$box <- result$box[ord,]
    result$box_metric <- result$box_metric[ord,]
    result$box_nom <- result$box_nom[ord]
  }
  return(result)
}

########################
### pasting function ###
########################
PRIM_paste <- function(fixbox, paste_alpha=0.01, max_steps=50, target=mean, stop_by_dec=T){
  data <- fixbox$data$all
  sub <- list(fixbox$data$box)
  f <-  fixbox$f
  beta <- fixbox$beta
  box_metric <- fixbox$fixbox_metric
  if(!is.null(box_metric)) row.names(box_metric) <- 1
  box_nom <- list(fixbox$fixbox_nom)
  metric_cov <- ifelse(!is.null(fixbox$fixbox_metric), ncol(fixbox$fixbox_metric)/2, 0)
  nom_var <- sapply(1:ncol(fixbox$data$all), function(k) is.factor(fixbox$data$all[,k]))
  for(i in 1:max_steps){
    if(!is.null(fixbox$fixbox_metric)){
      help_box <- lapply(1:metric_cov, function(k) box_metric[i,])
      help_box <- lapply(1:metric_cov, function(k) {help_box[[k]][1,k] <- -Inf; help_box[[k]]})
      lower_boxes <- lapply(1:metric_cov, function(k) {
        b <- box_metric[i,]
        b[1,k] <- next_value(vec=inbox(data, help_box[[k]], box_nom[[length(box_nom)]])[, !nom_var, drop=F][,k+1], border=as.numeric(box_metric[i,k]), N=ceiling(paste_alpha*nrow(sub[[i]])), lower=T)
        b
      })
      subsets_low <- lapply(1:metric_cov, function(k) inbox(data, lower_boxes[[k]], box_nom[[length(box_nom)]]))
      help_box <- lapply(1:metric_cov, function(k) box_metric[i,])
      help_box <- lapply(1:metric_cov, function(k) {help_box[[k]][metric_cov+k] <- Inf; help_box[[k]]})
      higher_boxes <- lapply(1:metric_cov, function(k) {
        b <- box_metric[i,]
        b[1,metric_cov+k] <- next_value(vec=inbox(data, help_box[[k]], box_nom[[length(box_nom)]])[, !nom_var, drop=F][,k+1], border=as.numeric(box_metric[i,metric_cov+k]), N=ceiling(paste_alpha*nrow(sub[[i]])), lower=F)
        b
      })
      subsets_high <- lapply(1:metric_cov, function(k) inbox(data, higher_boxes[[k]], box_nom[[length(box_nom)]]))
    } else subsets_low <- subsets_high <- NULL
    nom_boxes <- NULL
    if(!is.null(fixbox$fixbox_nom)){
      for(j in 1:length(box_nom[[i]])){
        nom_boxes <- c(nom_boxes, lapply(1:length(box_nom[[i]][[j]]), function(k) {a<-box_nom[[i]]; a[[j]][k]<-T; a}))
      }
      subsets_nom <- lapply(1:length(nom_boxes), function(k) inbox(data, box_metric[i,], nom_boxes[[k]]))
    } else subsets_nom <- NULL
    subsets <- c(subsets_low, subsets_high, subsets_nom)
    if(!is.null(fixbox$fixbox_metric)) b <- c(lower_boxes, higher_boxes)
    # evaluate the target function on all data sets
    t <- sapply(subsets, function(k) target(k[,1]))
    # if one box limit is already infinite, the target function therefor is set to NA
    # in this case no further data point could be added
    if(!is.null(fixbox$fixbox_metric)) t[1:ncol(box_metric)][!is.finite(as.numeric(box_metric[i,]))] <- NA
    # if one category of a nominal variable is already included, the target function of this subset becomes NA
    t[((metric_cov*2)+1):length(t)][unlist(box_nom[[length(box_nom)]]) == T] <- NA
    # find out which data set maximizes the target function
    which_max <- which.max(t)
    # end the loop if ....
    if(length(which_max) == 0) break
    if(stop_by_dec & t[which_max] < f[i]) break
    # update the next value of all outputs
    sub[[i+1]] <- subsets[[which_max]]
    f[i+1] <- t[which_max]
    beta[i+1] <- nrow(sub[[i+1]])/nrow(data)
    if (metric_cov!=0) if(which_max %in% 1:(metric_cov*2)) box_metric[i+1,] <- b[[which_max]] else box_metric[i+1,] <- box_metric[i,]
    if(which_max %in% ((metric_cov*2)+1):length(t)) box_nom[[i+1]] <- nom_boxes[[which_max-(metric_cov*2)]] else box_nom[[i+1]] <- box_nom[[length(box_nom)]]
  }
  if(!is.null(fixbox$fixbox_nom)){
    b_n <- as.data.frame(matrix(unlist(box_nom), ncol = length(unlist(box_nom[[1]])), byrow = T))
    names(b_n) <- names(unlist(box_nom[[1]]))
  } else b_n <- NULL
  if((!is.null(box_metric)) & (!is.null(b_n))) box <- cbind(box_metric, b_n) else 
    if(!is.null(box_metric)) box <- box_metric else box <- b_n
  return(list(f=f, beta=beta, box=box, box_metric=box_metric, box_nom=box_nom, data=sub, data_orig=fixbox$data$all))
}

########################
### peeling function ###
########################
PRIM_peel <- function(formula, data, peel_alpha= .05, beta_min= .01, target=mean, alter_crit=TRUE){
  # generate the data.frame X the function is working with
  if(!missing(formula)) X <- model.frame(formula, data=data) else X <- data
  # mark the ordinal covariates and set them into numeric 
  ord_cov <- sapply(1:ncol(X),function(k) is.ordered(X[,k]))[-1]
  for (i in 1:ncol(X)) if (is.ordered(X[,i])) X[,i] <- as.numeric(X[,i])
  # mark nominal variables
  nom_var <- sapply(1:ncol(X),function(k) is.factor(X[,k]))
  # which covartiates from all non nominal covariates are ordinal
  ord_cov_short <- ord_cov[!nom_var[-1]]
  # define start values
  beta <- 1
  f <- target(X[,1])
  sub <- list(X)
  crit <- NULL
  obs_in_box <- list(rep(TRUE, times=nrow(X)))
  if(sum(!nom_var[-1])>0){
    box <- rbind.data.frame(c(rep(-Inf, times=sum(!nom_var[-1])), rep(Inf, times=sum(!nom_var[-1]))))
    names(box) <- paste(rep(c("min", "max"), each=ncol(X[,!nom_var])-1), names(X[,!nom_var])[-1], sep=".")
  }
  if (sum(nom_var)>0){
    b_n <- rbind.data.frame(rep(TRUE, times=sum(sapply(1:ncol(subset(X, select=nom_var)), function(k) length(levels(subset(X, select=nom_var)[,k]))))))
    names(b_n) <- paste(rep(names(subset(X, select=nom_var)), times=sapply(1:ncol(subset(X, select=nom_var)), function(k) length(levels(subset(X, select=nom_var)[,k])))), unlist(lapply(1:ncol(subset(X, select=nom_var)), function(k) levels(subset(X, select=nom_var)[,k]))), sep=".")
    if(sum(!nom_var[-1])>0) box <- cbind(box, b_n) else box <- b_n
  }
  # loop 
  i <- 1
  while(TRUE){
    # generate all candidate subsets resulting from the ordinal and skalar variables
    if(sum(!nom_var[-1])>0){
      sub_metric <- sub[[i]][,!nom_var][,-1]
      box_mins <- apply(sub_metric, 2, function(k) quantile(k, probs = peel_alpha))
      box_maxs <- apply(sub_metric, 2, function(k) quantile(k, probs = 1 - peel_alpha))
      subsets_low <- lapply(1:length(box_mins), function(k) sub_metric[,k] > box_mins[k])
      subsets_high <- lapply(1:length(box_maxs), function(k)  subset = sub_metric[,k] < box_maxs[k])
    } else subsets_low <- subsets_high <- NULL
    # generate all candidate subsets resulting from the nominal variables
    subsets_nom <- NULL
    if (sum(nom_var)>0){
      sub_nom <- subset(sub[[i]], select=nom_var)
      for (j in 1:sum(nom_var)){
        subsets_nom <- c(subsets_nom, lapply(levels(sub_nom[,j]), function(k)  sub_nom[,j]!=k))
      }
    }
    # put all candidate subsets in one list object
    subsets <- c(subsets_low, subsets_high, subsets_nom)
    # put all possible updates of the box limits in one list
    if (sum(nom_var)>0) nom_update <- as.list(rep(FALSE, times=ncol(b_n))) else nom_update <- NULL
    if(sum(!nom_var[-1])>0) update <- as.list(c(box_mins, box_maxs)) else update <- NULL
    b <- c(update, nom_update)
    # evaluate the target function on all data sets
    t <- sapply(subsets, function(k) target(sub[[i]][,1][k]))
    # if one category of a nominal variable is already removed, the target function of this subset becomes NA
    t[box[i,] == FALSE & sapply(1:ncol(box), function(k) is.logical(box[,k]))] <- NA
    # alternative criterion
    if(alter_crit) cr <- (t - f[i]) / (nrow(sub[[i]]) - sapply(subsets, sum)) else cr <- t
    # find out which data set maximizes the target function
    which_max <- which.max(cr)
    # end the loop if ....
    if(length(which_max) == 0) break
    if(sum(subsets[[which_max]])/nrow(X) < beta_min) break
    # update the next value of all outputs
    obs_in_box[[i+1]] <- obs_in_box[[i]]
    obs_in_box[[i+1]][obs_in_box[[i+1]]]  <- subsets[[which_max]]
    sub[[i+1]] <- subset(sub[[i]], subset = subsets[[which_max]])
    f[i+1] <- t[which_max]
    crit[i+1] <- cr[which_max]
    beta[i+1] <- nrow(sub[[i+1]])/nrow(X)
    box[i+1,] <- box[i,]
    box[i+1, which_max] <- b[[which_max]]
    i<-i+1
  }
  # construct a data frame, which contains only the box limits of the "metric" (e.g. skalar or ordinal) Variables
  if(sum(!nom_var[-1])>0){
    box_metric <- box[, 1:(2*sum(!nom_var[-1]))]
  } else box_metric <- NULL
  # construct a list, which contains only the box limits of the nominal Variables
  if (sum(nom_var)>0){
    if(sum(!nom_var[-1])>0) box_n <- box[, -(1:(2*sum(!nom_var[-1])))] else box_n <- box
    lev <-  c(1, sapply(1:sum(nom_var), function(k) length(levels(subset(X, select=nom_var)[,k]))))
    box_nom <- lapply(1:nrow(box_n), function(k) lapply(1:(length(lev)-1), function(l) setNames(as.vector(t(box_n[k, sum(lev[1:l]):(sum(lev[1:(l+1)])-1) ])), levels(subset(X, select=nom_var)[,l]))))
    for(j in 1:length(box_nom)) names(box_nom[[j]]) <- names(X)[nom_var]
  } else box_nom <- NULL
  return(list(f=f, beta=beta, box=box, box_metric=box_metric, box_nom=box_nom, data=sub, data_orig=X, crit=crit, obs_in_box=obs_in_box))
}

###########################################
### peeling function with bootstrapping ###
###########################################
PRIM_peel_bs <- function(formula, data, target=mean, B=5, peel_alpha=seq(0.01, 0.5, 0.01), seed, beta_min= .01, alter_crit=TRUE, print_position=TRUE){
  # generate the data.frame X the function is working with
  if(!missing(formula)) X <- model.frame(formula, data=data) else X <- data
  if(!missing(seed)) set.seed(seed)
  box <- NULL
  box_metric <- NULL
  box_nom <- NULL
  # generate B bootstrap samples for each alpha 
  # and evaluate the PRIM function on each to get the box limits
  for(i in peel_alpha){
    if(print_position) cat("\nalpha =", i, "\n Iteration: " )
    for (j in 1:B){
      if(j==1) X_new <- X else X_new <- X[sample(1:nrow(X), replace = TRUE), ]
      prim <- PRIM_peel(data=X_new, peel_alpha=i, beta_min = beta_min, alter_crit = alter_crit, target = target)
      box <- rbind(box, prim$box)
      box_metric <- rbind(box_metric, prim$box_metric)
      if(!is.null(prim$box_nom)) box_nom <- c(box_nom, prim$box_nom)
      if(print_position) cat(j, " ")
    }
  }
  names(box) <- names(prim$box)
  for (i in 1:ncol(X)) if (is.ordered(X[,i])) X[,i] <- as.numeric(X[,i])
  sub <- lapply(1:nrow(box_metric), function(k) {
    inbox(X, fixbox_metric=box_metric[k,], fixbox_nom=box_nom[[k]])
  })
  f <- sapply(sub, function(k) target(k[,1]))
  beta <- sapply(1:length(f), function(k) nrow(sub[[k]])/nrow(X))
  return(list(f=f, beta=beta, box=box, box_metric=box_metric, box_nom=box_nom, data=sub, data_orig=X))
}

###############################
### PRIM with bootstrapping ###
###############################
PRIM_bs <- function(formula, data, beta_min = 0.1, f_min, target=mean, B=10, peel_alpha=seq(0.01, 0.1, 0.01), seed, alter_crit=T,
                    paste_alpha=0.01, max_steps=50, stop_by_dec=T, print_position=T){
  fixboxes <- list(NULL)
  i <- 1
  remove_obs <- list(NULL)
  while(TRUE){
    if(print_position) cat("\n\nBox", i)
    if(i==1) {
      peel <- PRIM_peel_bs(formula = formula, data = data, B = B, peel_alpha = peel_alpha, seed = seed, beta_min = beta_min, alter_crit = alter_crit, print_position = print_position) 
    } else {
      peel <- PRIM_peel_bs(formula = formula, data = data[-remove_obs[[i]], ], B = B, peel_alpha = peel_alpha, seed = seed, beta_min = beta_min, alter_crit = alter_crit, print_position = print_position)
    }
    if(i==1) {
      peel$f[peel$beta < beta_min] <- NA
      peel$beta[peel$beta < beta_min] <- NA
    } else {
      peel$f[peel$beta < (beta_min*nrow(data)/nrow(data[-remove_obs[[i]], ]))] <- NA
      peel$beta[peel$beta < (beta_min*nrow(data)/nrow(data[-remove_obs[[i]], ]))] <- NA
    }
    if(!missing(f_min)) {
      peel$beta[peel$f < f_min] <- NA
      which_max_peel <- which.max(peel$beta)
    } else which_max_peel <- which.max(peel$f)
    if(length(which_max_peel) == 0) which_max_peel <- 1
    fixbox_peel <- define_fixbox(peel, which_max_peel)
    Paste <- PRIM_paste(fixbox = fixbox_peel, paste_alpha = paste_alpha, max_steps = max_steps, stop_by_dec = stop_by_dec)
    which_max_paste <- which.max(Paste$f)
    fixboxes[[i]] <- define_fixbox(Paste, which_max_paste)
    remove_obs[[i+1]] <- c(remove_obs[[i]], as.numeric(row.names(fixboxes[[i]]$data$box)))
    if(fixboxes[[i]]$f <= target(fixboxes[[1]]$data$all[,1]) ) break
    i <- i+1
  }
  f  <- box <- sub <- box_metric <- box_nom <- NULL
  if(length(fixboxes)>1){
    for(i in 1:(length(fixboxes)-1)){
      f <- c(f,fixboxes[[i]]$f)
      box <- rbind(box,fixboxes[[i]]$fixbox)
      if(!is.null(fixboxes[[1]]$fixbox_metric)) box_metric <- rbind(box_metric,fixboxes[[i]]$fixbox_metric)
      if(!is.null(fixboxes[[1]]$fixbox_nom)) box_nom <- c(box_nom, list(fixboxes[[i]]$fixbox_nom))
      sub <- c(sub, list(fixboxes[[i]]$data$box))
    }
  } else {
    ret <- list(f=f, beta=beta, box=box, box_metric=box_metric, box_nom=box_nom, data=list(data))
    class(ret) <- "PRIM"
    return(ret)
  }
  sub[[length(sub)+1]] <- fixboxes[[length(fixboxes)]]$data$all
  f[[length(f)+1]] <- target(sub[[length(sub)]][,1])
  beta <- sapply(1:length(sub), function(k) nrow(sub[[k]])/nrow(data))
  ret <- list(f=f, beta=beta, box=box, box_metric=box_metric, box_nom=box_nom, data=sub)
  class(ret) <- "PRIM"
  return(ret)
}

########################
### predict function ###
########################
predict.PRIM <- function(p, type = c("response", "box")[1]) {
  output <- rep(NA, sum(sapply(p$data, nrow)))
  for (i in 1:length(p$data)) {
    if (type == "box") output[as.numeric(rownames(p$data[[i]]))] <- i
    if (type == "response") output[as.numeric(rownames(p$data[[i]]))] <- mean(p$data[[i]][, 1])
  }
  output
}


