##########################
### required libraries ###
##########################
library(mfp)
library(rseg)
library(survival)
library(rms)

######################
### Example of use ###
######################

# Load the German Breast Cancer Study Group data
data(GBSG)

# seed for reproducibility
set.seed(1234) 

# train CART based rseg model
seg  <-  rSeg(Surv(rfst, cens) ~ age + menostat + tumsize + tumgrad + posnodal + prm + esm, data = GBSG[1:480, ])

# plot rseg and rpart models
plot(gettree(seg))
plot(seg)

# record allocation of observations to nodes and segments
GBSG$node <- predict(gettree(seg), newdata = GBSG, type = "node")
GBSG$segment <- predict(seg, newdata = GBSG, type = "segment")

# plots of rpart model on training an test data
survplot(npsurv(Surv(rfst, cens) ~ node, data = GBSG[1:480, ]), lty = 1, type = "kaplan-meier", conf.type = "log-log", col.fill = rep(adjustcolor("grey", alpha.f = 0.4), 11), ylab = "recurrence-free survival", xlab = "days")

survplot(npsurv(Surv(rfst, cens) ~ node, data = GBSG[481:686, ]), lty = 1, type = "kaplan-meier", conf.type = "log-log", col.fill = rep(adjustcolor("grey", alpha.f = 0.4), 11), ylab = "recurrence-free survival", xlab = "days")

# plots of rseg model on training an test data
survplot(npsurv(Surv(rfst, cens) ~ segment, data = GBSG[1:480, ]), lty = 1, type = "kaplan-meier", conf.type = "log-log", col.fill = rep(adjustcolor("grey", alpha.f = 0.4), 3), ylab = "recurrence-free survival", xlab = "days")

survplot(npsurv(Surv(rfst, cens) ~ segment, data = GBSG[481:686, ]), lty = 1, type = "kaplan-meier", conf.type = "log-log", col.fill = rep(adjustcolor("grey", alpha.f = 0.4), 3), ylab = "recurrence-free survival", xlab = "days")








